;; Calculate cross sections according to the equations by Volker
;; Ossenkopf.  It writes a file which contains the cross sections for
;; collisions with gas, and a matrix with the cross sections for
;; collisions between grains of different size.


;; The procedure takes the following keyword parameters:
;;
;; ptype     - The particle type, one of COMPACT, PCA, CCA
;; amin_DIS  - The minimum radius of the initial size distribution
;;             which.  All these particles are assumed to be compact
;; amax_DIS  - The maximum radius of the initial size distribution
;;             which.  All these particles are assumed to be compact
;; p_DIS     - The powerlaw of the initial size distribution
;; mmin      - Minimum mass of the mass grid
;; mmax      - Maximum mass of the mass grid
;; Nmass     - Number of mass bins
;; mass      - If defined, then it will take this mass grid instead of own
;; rhodust   - Specific density of compact particles
;; showplots - When set, show the computed cross sections as plots


pro crosssections,ptype=ptype, $
                   amin_DIS=amin_DIS,amax_DIS=amax_DIS, p_DIS=p_DIS,$
                   mmin=mmin,mmax=mmax,Nmass=Nmass, $
                   rhodust=rhodust,mass=mass,$
                   showplots=showplots
                   

  if not keyword_set(ptype) then begin
      print,"Particle type not specified, using COMPACT"
      ptype = 'COMPACT'
  endif
  if not (ptype eq 'COMPACT' or ptype eq 'PCA' or PTYPE eq 'CCA') then begin
      print,'ERROR: Unknown particle type ',ptype
      stop
  end 
  ;; The size limits of the monomer size distribution
  micron = 1d-4
  if not keyword_set(amin_DIS) then amin_DIS=0.01*micron
  if not keyword_set(amax_DIS) then amax_DIS=0.5*micron
  if not keyword_set(p_DIS) then p_DIS = -3.5

  ;; The density of the grain material in the monomer distribution
  if not keyword_set(rhodust) then rhodust = 3.7     ; g/cm^3

  ;; The mass limits of the grid. Default from amin_DIS and a 1km body
  if not keyword_set(mmin) then mmin = 4.d0*!DPI/3.d0*amin_DIS^3*rhodust 
  if not keyword_set(mmax) then mmax = 4.d0*!DPI/3.d0*1d5^3*rhodust 
  if not keyword_set(Nmass) then Nmass = 100
  if nmass ge 1000 then begin
      print,'ERROR: Too many mass bins ',nmass
      stop
  endif 

  ;; Parameters for the input size distribution (MRN like)
  a1_dis = amin_DIS
  a2_dis = amax_DIS
  p_dis = -3.5

  ;; calculate the mean radius, area, volume, mass of a grain
  a_dis_average = dblarr(4)
  a_dis_average[0] = 1.d0;
  for i=1,3 do begin
      p1 = p_dis+i+1
      p2 = p_dis+1
      a_dis_average[i] = (p2/p1 * (a2_dis^p1 - a1_dis^p1) / $
                         (a2_dis^p2 - a1_dis^p2))^(1.d0/i)
  endfor
  meanrad_dis = a_dis_average[1]
  meansig_dis = !DPI*a_dis_average[2]^2
  meanvol_dis = 4.d0/3.d0*!DPI*a_dis_average[3]^3
  meanmass_dis = meanvol_dis * rhodust
  print,'Mean sizes',meanrad_dis,meansig_dis,meanvol_dis,meanmass_dis

  ;; parameters for the mass distribution in the table
  if n_elements(mass) eq 0 then begin
     m1 = mmin
     m2 = mmax
     logmass = dindgen(nmass)/(nmass-1.d0) * (alog10(m2)-alog10(m1))+alog10(m1)
     mass = 10.d0^logmass
  endif else begin
     nmass = n_elements(mass)
  endelse

  ;; Compute the compact volume of the grains
  volume = mass/rhodust
  ;; Compute volume-equivalent radius
  radius = (3.d0*mass/(4.d0*!DPI*rhodust))^(1.d0/3.d0)
  ;; Compute the normalized volume with respect to mean distribution volume
  V = volume / meanvol_dis
  Atilde = V

  if ptype eq 'CCA' then begin
      ;; CCA particles
      a0 = 0.692 & b =0.301
      w = where(V ge 25.d0)
      Atilde[w] = V[w]^(-0.05d0) * a0 * (1.d0+b/alog(V[w]))
      ;; small are treated like PCA
      a0 = 15.2 & b = 2.86 & gamma = 0.096
      w = where(V lt 25.d0)
      Atilde[w] = V[w]^(-1.d0/3.) * a0 * exp(-b/V[w]^gamma)
  endif else if ptype eq 'PCA' then begin
      ;; PCA particles
      a0 = 15.2d0 & b = 2.86d0 & gamma = 0.096d0
      w = where(V ge 2.d0 and V lt 3000.d0)
      Atilde[w] = V[w]^(-1.d0/3.) * a0 * exp(-b/V[w]^gamma)
      a0 = 4.70 & b = 9.02 & gamma = 0.503d0
      w = where(V ge 600.d0)
      Atilde[w] = V[w]^(-1.d0/3.) * a0 * exp(-b/V[w]^gamma)
  endif else begin
      ;; Compact particles
      Atilde = V^(-1.d0/3.d0)
  endelse
  ;; Correction: For particles less than 2, use compact properties
  w = where(V lt 2.d0)
  if n_elements(w) gt 1 then begin
      Atilde[w] = V[w]^(-1.d0/3.d0)
  endif

  if ptype eq 'CCA' or ptype eq 'PCA' then begin

      ;; Calculate the dimensionless A
      A = Atilde * V
      
      ;; Calculate the real A in physical units
      Aphys = A * (meanvol_dis*3.d0/(4.d0*!DPI))^(2./3.)*!DPI
      
      ;; Fluffyness parameter
      x = Atilde^3*V
      
      ;; toothing radius
      a2 = 0.720  & b = 0.216
      atooth = a2 * V^(1.d0/3.d0) * x^0.21 * (1.d0-b/x^1.d0/3.d0)
      plot,V,atooth,/xlog,/ylog,xrange=[1,10000]

      ;; Collision cross sections
      sigma_coll = dblarr(nmass,nmass)
      a0 = 8.3d0 & d2 = 1.22d0
      for j1=0,nmass-1 do begin
          for j2 = 0,nmass-1 do begin
              ;; Sort the indices correctly
              if atooth[j1] ge atooth[j2] then begin
                  i1=j1 & i2=j2
              endif else begin
                  i1=j2 & i2=j1
              endelse 
              sigma_coll[i1,i2] = $
                A[i1] * $
                (1.d0-(atooth[i2]/atooth[i1])^2) + $
                (4.d0*!DPI - $
                 a0 * (1.d0-(atooth[i1]/atooth[i2])^d2)) * atooth[i2]^2
              sigma_coll[i1,i2] = sigma_coll[i1,i2] * meansig_dis
              sigma_coll[i2,i1] = sigma_coll[i1,i2]
          endfor 
      endfor
  endif else begin
      ;; Compact particles
      atooth=radius
      x = radius*0.d0
      Aphys = !DPI*radius^2
      sigma_coll = dblarr(nmass,nmass)
      for i1=0,nmass-1 do begin
          for i2 = 0,nmass-1 do begin
              sigma_coll[i1,i2] = !DPI*(radius[i1]+radius[i2])^2
          endfor 
      endfor
  endelse 

  if keyword_set(showplots) then begin
      ;; Plot the gas cross sections
      window,0
      plot,mass,Aphys,/xlog,/ylog, $
           xtitle='mass',ytitle='gas crossection', $
           title='Cross section for collisions with gas atoms'
      ;; For comparison, plot the compact particles
      oplot,mass,!DPI*radius^2,line=2

      
      ;; Plot the collisional cross sections
      window,1
      levels = indgen(100)-50
      contour,alog10(sigma_coll),mass,mass, $
              /xlog,/ylog,levels=congrid(levels,60),c_labels=congrid(levels,60)+1000, $
              xtitle='mass',ytitle='mass', $
              title='Cross section for collisions between grains'
  endif  
  

  ;; Write the files
  openw,1,'amgrid.inp'
  printf,1,nmass
  printf,1,' '
  for i=0,nmass-1 do begin
     printf,1,sqrt(Aphys[i]/!dpi),mass[i]
  endfor
  close,1
  openw,1,'coagcrosssec.inp'
  printf,1,nmass
  printf,1,' '
  for i=0,nmass-1 do begin
     for k=0,nmass-1 do begin
        printf,1,sigma_coll[i,k]
     endfor
  endfor
  close,1
end
