!======================================================================
!          1+1 D Disk structure code with dust settling/mixing
!                          C.P. Dullemond
!                            July 2009
!                Modified July 2010 (drag law improved)
!======================================================================

!----------------------------------------------------------------------
!                             MAIN PROGRAM
!----------------------------------------------------------------------
program main
  use userdef_module
  use variables_module
  implicit none
  double precision :: GG,kk,mp,year
  parameter(GG=6.673d-8,mp=1.6726d-24,kk=1.3807d-16,year=3.15576d7)
  double precision :: tempg, rhog, force, om2, omk, vkep, cs2, dt0
  double precision :: sc, st, timecur, timenext, vmax, vdrift
  double precision, allocatable :: zero(:),rhodbk(:,:),rhogbk(:,:)
  integer :: itime,ir,iz,is,number_invalid,itt
  logical :: conv,fex
  character*80 base,ext,filename
  double precision :: cst
  parameter(cst=2.1277d0)        ! 8*sqrt(2)/(3*sqrt(pi))
  !
  ! Set default parameters
  !
  driftold         = .false.
  star_m           = 1.98892d33
  time_now         = 0.d0
  time_end         = 3d7
  time_dtsave      = -1
  time_dtt         = 0.d0
  dtomk            = 0.2 
  mugas            = 2.3
  vmaxfrac         = 0.5
  schmidt_powerlaw = 1.d0            ! The classical definition of Schmidt nr
  !schmidt_powerlaw = 2.d0            ! The Youdin & Lithwich definition of Schmidt nr
  steady_tolerance = 1d-6
  incl_settling    = 1
  incl_mixing      = 1
  nstepsave        = 1000000
  maxnrsave        = 1000
  !
  ! Set defaults for user defined variables
  !
  call userdef_set_defaults()
  !
  ! Read the diskmixsett input file
  !
  call read_diskmixsett_file()
  !
  ! Consistency checks
  ! 
  if(star_m.lt.0.d0) then
     write(*,*) 'ERROR: Must set star_m in diskmixsett.inp file.'
     stop
  endif
  if(driftold) then
     write(*,*) '************************************************'
     write(*,*) ' WARNING: Using old Dullemond & Dominik 2004'
     write(*,*) '     drag law, which is not correct. Apparently '
     write(*,*) '     you have set drift_old_formula=1 in the'
     write(*,*) '     diskmixsett.inp file.'
     write(*,*) '************************************************'
  endif
  !
  ! Check if a savetimes.inp file is present
  !
  if(allocated(savetimes_time)) deallocate(savetimes_time)
  inquire(file='savetimes.inp',exist=fex)
  if(fex) then
     if(time_dtsave.gt.0.d0) then
        write(*,*) 'ERROR: Found savetimes.inp file, so the dump times'
        write(*,*) '       are now taken from that file. Must put'
        write(*,*) '       time_dtsave to 0 or not specify i diskmixsett.inp.'
        stop 
     endif
     open(unit=1,file='savetimes.inp')
     read(1,*) savetimes_nr
     if(savetimes_nr.lt.1) stop
     allocate(savetimes_time(1:savetimes_nr))
     read(1,*) savetimes_time
     close(1)
  else
     if(time_dtsave.eq.0.d0) then
        write(*,*) 'ERROR: Found time_dtsave==0.d0. Seems that you are'
        write(*,*) '       expecting the saving times from savetimes.inp.'
        write(*,*) '       But no such file found...'
        stop 
     endif
     savetimes_nr=0
     if(time_dtsave.le.0.d0) time_dtsave = time_end
  endif
  !
  ! Read the main disk struct file, which contains the gas density and
  ! gas temperature
  !
  ! NOTE: This file has the format of the input files from the 
  !       diskstruct code, but without the dust stuff. So only
  !       the basic five columns: r,z,rho,temp,dummy. The dust and
  !       chemical species we will do in a separate file in this
  !       code.
  !
  ! NOTE: This automatically also defines the r and z grids.
  !
  call read_diskstruct_file()
  !
  ! Read auxiliary variables, such as the turbalpha(R,z). Depends on the 
  ! iformat number which variables are listed here. For a start we just
  ! put turbalpha(r,z) here.
  !
  call read_auxvar_file()
  !
  ! Read the current vertical distribution of the dust, as well as some
  ! properties of each dust component.
  !
  call read_dustdens_file()
  !
  ! Now read the gas chemical species
  !
  call read_chemdens_file()
  !
  ! Now call the userdef reading routine, if required
  !
  call userdef_readall()
  !
  ! Allocate some arrays
  !
  if(allocated(rhodust)) then
     allocate(dust_vz(1:grid_nz+1,1:dust_nrspec,1:grid_nr))
     allocate(diffcoefdust(1:grid_nz,1:dust_nrspec,1:grid_nr))
     allocate(rhodbk(1:grid_nz,1:dust_nrspec))
  endif
  if(allocated(rhochem)) then
     allocate(rhogbk(1:grid_nz,1:chem_nrspec))
  endif
  allocate(diffcoefgas(1:grid_nz,1:grid_nr))
  allocate(zero(1:grid_nz))
  !
  ! Calculate the settling velocity for all dust species at all heights
  ! and radii
  !
  if(allocated(rhodust)) then
     if(incl_settling.eq.1) then
        do ir=1,grid_nr
           om2    = GG * star_m / grid_r(ir)**3
           vkep   = sqrt( GG * star_m / grid_r(ir) )
           do is=1,dust_nrspec
              dust_vz(1,is,ir) = 0.d0
              dust_vz(grid_nz+1,is,ir) = 0.d0
              do iz=2,grid_nz
                 !
                 ! Calculate the vertical drift speed of this dust size 
                 ! at this location
                 !
                 tempg             = 0.5d0 * ( tgas(iz-1,ir) + tgas(iz,ir) )
                 rhog              = 0.5d0 * ( rhogas(iz-1,ir) + rhogas(iz,ir) )
                 force             = dust_m(is)*om2*grid_zi(iz,ir)
                 dust_vz(iz,is,ir) = -vdrift(tempg,rhog,mugas,force,dust_sigfr(is),driftold)
                 !
                 ! Check the validity of the vdrift
                 !
                 if(number_invalid(dust_vz(iz,is,ir)).ne.0) then
                    write(*,*) 'ERROR: NaN or Inf value detected (2)'
                    write(*,*) '   iz,is,ir   = ',iz,is,ir
                    write(*,*) '   vz         = ',dust_vz(iz,is,ir)
                    stop 13
                 endif
                 !
                 ! But it is not allowed to exceed (z/R) v_kepler
                 !
                 vmax = vmaxfrac * vkep * grid_zi(iz,ir) / grid_r(ir)
                 dust_vz(iz,is,ir) = - ( abs(dust_vz(iz,is,ir))**(-1.) +  &
                      vmax**(-1.) )**(-1.)
              enddo
           enddo
        enddo
     else
        dust_vz(:,:,:) = 0.d0
     endif
  endif
  !
  ! Calculate the gas diffusion coefficient
  !
  if(incl_mixing.eq.1) then
     do ir=1,grid_nr
        omk    = sqrt( GG * star_m / grid_r(ir)**3 )
        do iz=1,grid_nz
           cs2 = kk * tgas(iz,ir) / ( mugas * mp )
           diffcoefgas(iz,ir) = turbalpha(iz,ir) * cs2 / omk
        enddo
     enddo
  else
     diffcoefgas(:,:) = 0.d0
  endif
  !
  ! Calculate the dust diffusion coefficient
  !
  if(allocated(rhodust)) then
     if(incl_mixing.eq.1) then
        do ir=1,grid_nr
           omk    = sqrt( GG * star_m / grid_r(ir)**3 )
           do is=1,dust_nrspec
              do iz=1,grid_nz
                 cs2 = kk * tgas(iz,ir) / ( mugas * mp )
                 !
                 ! Compute the Stokes number according to Eq. 15 of
                 ! Dullemond & Dominik 2004b
                 !
                 ! NOTE: We have chosen the q-parameter to be q=0.5,
                 !       as is suggested in the DD04b paper.
                 ! 
                 ! NOTE: With the improved drag law (Schaaf 1963)
                 !       also this part changes.
                 !
                 if(driftold) then
                    !
                    ! Using the original (wrong) Dullemond & Dominik (2004)
                    ! formula for the Epstein drag law.
                    !
                    st = (3.d0/4.d0) * (dust_m(is)/dust_sigfr(is)) *     &
                         omk / (rhogas(iz,ir)*sqrt(cs2))
                 else
                    !
                    ! Using the correct Schaaf (1963) formula for the
                    ! Epstein drag law.
                    !
                    st = (1.d0/cst) * (dust_m(is)/dust_sigfr(is)) *     &
                         omk / (rhogas(iz,ir)*sqrt(cs2))
                 endif
                 !
                 ! Compute the Schmidt number
                 !
                 ! NOTE: The schmidt_powerlaw thingie is to switch between
                 !       the classical Sc = 1 + St to the Youdin & Lithwick
                 !       Sc = 1 + St^2 new formulation.
                 !
                 sc = 1.d0 + st**schmidt_powerlaw
                 !
                 ! Now compute the dust diffusion coefficient
                 !
                 diffcoefdust(iz,is,ir) = diffcoefgas(iz,ir) / sc
              enddo
           enddo
        enddo
     else
        diffcoefdust(:,:,:) = 0.d0
     endif
  endif
  !
  ! Prepare some output files
  !
  open(unit=1,file='time.out')
  write(1,*) ' '
  close(1)
  itime = 0
  open(unit=1,file='time.info')
  write(1,*) itime
  close(1)
  !
  ! Do some last-minute checks for the user defined functions
  !
  call userdef_checkall()
  !
  ! Write the starting state
  !
  write(*,*) 'Writing dump ',itime,' at time ',time_now/year,' years'
  open(unit=1,file='time.info')
  write(1,*) itime
  close(1)
  open(unit=1,file='time.out',access='append')
  write(1,*) time_now
  close(1)
  if(allocated(rhodust)) call write_dustdens_file(itime)
  if(allocated(rhochem)) call write_chemdens_file(itime)
  call userdef_writeframe(itime)
  !
  ! Finally do the actual computational loop. Do this from one dump
  ! to the next
  !
  do while(time_now.lt.time_end)
     !
     ! Compute the next dump time
     !
     if(allocated(savetimes_time)) then
        if(itime+1.le.savetimes_nr) then
           time_nextdump = savetimes_time(itime+1)
        else
           time_nextdump = time_end
        endif
     else
        time_nextdump = time_now + time_dtsave
     endif
     !
     ! Check that it does not exceed time_end
     !
     if(time_nextdump.gt.time_end) time_nextdump = time_end
     !
     ! Do a loop over radius, i.e. advect each radius separately,
     ! because they all have their own time scale. Some radii may 
     ! reach quickly a steady-state, and we do not need to model
     ! further, while others may take longer to reach a steady state
     ! or the next dump time.
     !
     do ir=1,grid_nr
        !
        ! Compute a reasonable time step. Strictly speaking the 
        ! implicit scheme can take large time steps, but to get
        ! accurate answers, we still should be careful not to 
        ! overdo it. 
        ! 
        omk     = sqrt( GG * star_m / grid_r(ir)**3 )
        dt0     = dtomk / omk
        !
        ! Reset the time
        !
        timecur = time_now
        !
        ! Now do the time sub-loop
        !
        itt = 1
        do while(timecur.lt.time_nextdump)
           !
           ! Compute the time step
           !
           if(time_dtt.gt.0.d0) then
              time_dt = max(time_dtt*timecur,dt0)
           else
              time_dt = dt0
           endif
           !
           ! Compute the next time
           !
           timenext = timecur + time_dt
           !
           ! Check that we don't exceed dump time
           !
           if(timenext.gt.time_nextdump) then
              time_dt  = time_nextdump - timecur
              timenext = time_nextdump
           endif
           !
           ! Are we exceeding max nr of time steps before saving?
           ! WARNING: For more than one radial grid point this may not
           !          work well at all times.
           !
           if(itt.ge.nstepsave) then
              time_nextdump = timenext
              itt = 0
           endif
           !
           ! Do the backup of the current dust and gas densities
           !
           if(allocated(rhodust)) rhodbk(:,:) = rhodust(:,:,ir)
           if(allocated(rhochem)) rhogbk(:,:) = rhochem(:,:,ir)
           !
           ! Do the advection-diffusion of the dust
           !
           if(allocated(rhodust)) then
              do is=1,dust_nrspec
                 call do_turb_diff_and_sett(grid_nz,grid_zi(1,ir),dust_vz(1,is,ir),  &
                      time_dt,rhogas(1,ir),diffcoefdust(1,is,ir),rhodust(1,is,ir))
              enddo
           endif
           !
           ! Do the diffusion of the gas chemical species
           !
           if(allocated(rhochem)) then
              do is=1,chem_nrspec
                 call do_turb_diff_and_sett(grid_nz,grid_zi(1,ir),zero,  &
                      time_dt,rhogas(1,ir),diffcoefgas(1,ir),rhochem(1,is,ir))
              enddo
           endif
           !
           ! Here you can add any reactions. Note that this will then 
           ! be done in an operator-splitting way, which is not always
           ! very trustworthy
           !
           call userdef_do_reactions()
           !
           ! Now check if we reached a steady state yet
           !
           conv = .true.
           do iz=1,grid_nz
              if(allocated(rhodust)) then
                 do is=1,dust_nrspec
                    if(rhodust(iz,is,ir).lt.0.d0) then
                       write(*,*) 'ERROR: Negative dust density found...'
                       write(*,*) iz,is,ir
                       stop
                    endif
                    if(abs(rhodust(iz,is,ir)/(rhodbk(iz,is)+1d-90)-1.d0)  &
                         .gt.steady_tolerance) then
                       conv=.false.
                    endif
                 enddo
              endif
              if(allocated(rhochem)) then
                 do is=1,chem_nrspec
                    if(rhochem(iz,is,ir).lt.0.d0) then
                       write(*,*) 'ERROR: Negative chem density found...'
                       write(*,*) iz,is,ir
                       stop
                    endif
                    if(abs(rhochem(iz,is,ir)/(rhogbk(iz,is)+1d-90)-1.d0)  &
                         .gt.steady_tolerance) then
                       conv=.false.
                    endif
                 enddo
              endif
           enddo
           !
           ! Update time
           !
           if(conv) then
              !
              ! We have locally reached a steady state: go to end of
              ! this time interval
              !
              timecur = time_nextdump
           else
              !
              ! Normal update of time
              !
              timecur = timenext
           endif
           itt = itt + 1
           !
        enddo
     enddo
     !
     ! Update time
     !
     time_now = time_nextdump
     !
     ! Now dump the current results
     !
     itime = itime+1
     write(*,*) 'Writing dump ',itime,' at time ',time_now/year,' years'
     open(unit=1,file='time.info')
     write(1,*) itime
     close(1)
     open(unit=1,file='time.out',access='append')
     write(1,*) time_now
     close(1)
     if(allocated(rhodust)) call write_dustdens_file(itime)
     if(allocated(rhochem)) call write_chemdens_file(itime)
     call userdef_writeframe(itime)
     !
     ! Check if we do not exceed max nr of saves (not to accidently
     ! flood the hard drive)
     ! 
     if(itime.ge.maxnrsave) then
        write(*,*) 'STOPPING: Max nr of saves reached...'
        stop
     endif
     !
  enddo
  !
  ! Check if there are any further output files from previous runs
  !
  base = 'dustdens_'
  ext  = '.out'
  call make_indexed_filename_filled(base,itime+1,ext,filename)
  inquire(file=filename,exist=fex)
  if(fex) then
     write(*,*) 'WARNING: Old dustdens_***.out files still remain: ', &
          trim(filename)
  endif
  base = 'chemdens_'
  ext  = '.out'
  call make_indexed_filename_filled(base,itime+1,ext,filename)
  inquire(file=filename,exist=fex)
  if(fex) then
     write(*,*) 'WARNING: Old chemdens_***.out files still remain: ', &
          trim(filename)
  endif
  !
  ! Cleanup
  !
  if(allocated(grid_r)) deallocate(grid_r)
  if(allocated(grid_z)) deallocate(grid_z)
  if(allocated(grid_zi)) deallocate(grid_zi)
  if(allocated(rhogas)) deallocate(rhogas)      
  if(allocated(tgas)) deallocate(tgas)        
  if(allocated(turbalpha)) deallocate(turbalpha)
  if(allocated(zero)) deallocate(zero)    
  if(allocated(rhodust)) deallocate(rhodust)
  if(allocated(rhochem)) deallocate(rhochem)
  if(allocated(rhodbk)) deallocate(rhodbk)
  if(allocated(rhogbk)) deallocate(rhogbk)
  if(allocated(diffcoefgas)) deallocate(diffcoefgas)
  if(allocated(diffcoefdust)) deallocate(diffcoefdust)
  if(allocated(savetimes_time)) deallocate(savetimes_time)
  call userdef_cleanup()
end program main


!-------------------------------------------------------------------------
!                      READ THE diskmixsett.inp FILE
!-------------------------------------------------------------------------
subroutine read_diskmixsett_file()
  use namelist_module
  use userdef_module
  use variables_module
  implicit none
  character*160 :: string
  logical :: fex1
  integer :: idum,len,idriftold
  logical :: ierror
  !
  inquire(file='diskmixsett.inp',exist=fex1)
  if(fex1) then 
     idriftold = -1
     open(unit=1,file='diskmixsett.inp')
     call read_input_file()
     close(1)
     call parse_input_double ('star_m@                       ',star_m)
     call parse_input_double ('time_end@                     ',time_end)
     call parse_input_double ('time_now@                     ',time_now)
     call parse_input_double ('time_dtsave@                  ',time_dtsave)
     call parse_input_double ('schmidt_powerlaw@             ',schmidt_powerlaw)
     call parse_input_double ('vmaxfrac@                     ',vmaxfrac)
     call parse_input_double ('dtomk@                        ',dtomk)
     call parse_input_double ('steady_tolerance@             ',steady_tolerance)
     call parse_input_double ('time_dtt@                     ',time_dtt)
     call parse_input_integer('incl_settling@                ',incl_settling)
     call parse_input_integer('incl_mixing@                  ',incl_mixing)
     call parse_input_integer('drift_old_formula@            ',idriftold)
     call parse_input_integer('nstepsave@                    ',nstepsave)
     call parse_input_integer('maxnrsave@                    ',maxnrsave)
     call userdef_parse_input()
     call check_all_lines_ok(ierror)
     if(ierror) stop
     if(idriftold.ge.0) then
        if(idriftold.eq.0) then
           driftold = .false.
        else
           driftold = .true.
        endif
     endif
  else
     write(*,*) 'ERROR: Could not find file diskmixsett.inp'
     stop
  endif
  !
end subroutine read_diskmixsett_file


!-------------------------------------------------------------------------
!                      READ THE diskstruct.inp FILE
!-------------------------------------------------------------------------
subroutine read_diskstruct_file()
  use variables_module
  implicit none
  integer :: ir,iz
  double precision :: a,b,c,d,e
  !
  ! Open the file
  !
  open(unit=1,file='diskstruct.inp',status='old')
  read(1,*) grid_nr,grid_nz
  if(grid_nr.lt.1) then
     write(*,*) 'ERROR: This program does not accept the multi-sizes/species'
     write(*,*) '       format of the diskstruct.inp file. All dust species'
     write(*,*) '       should be in diskdust.inp and all gas chemical'
     write(*,*) '       species in chemspec.inp'
     stop
  endif
  if(grid_nz.lt.1) stop 890
  !
  ! Allocate the grid arrays
  !
  allocate(grid_r(1:grid_nr))
  allocate(grid_z(1:grid_nz,1:grid_nr))
  allocate(grid_zi(1:grid_nz+1,1:grid_nr))
  allocate(rhogas(1:grid_nz,1:grid_nr))
  allocate(tgas(1:grid_nz,1:grid_nr))
  !
  ! Read the stuff
  !
  do ir=1,grid_nr
     do iz=1,grid_nz
        read(1,*) a,b,c,d,e
        grid_r(ir) = a
        grid_z(iz,ir) = b
        tgas(iz,ir) = d
        rhogas(iz,ir) = c 
     enddo
  enddo
  close(1)
  !
  ! Create the zi grid
  !
  do ir=1,grid_nr
     grid_zi(1,ir) = 0.d0
     do iz=2,grid_nz
        grid_zi(iz,ir) = 0.5d0 * ( grid_z(iz,ir) + grid_z(iz-1,ir) )
     enddo
     grid_zi(grid_nz+1,ir) = 2*grid_zi(grid_nz,ir) - grid_zi(grid_nz-1,ir) 
  enddo
  !  
end subroutine read_diskstruct_file


!-------------------------------------------------------------------------
!                      READ THE auxvar.inp FILE
!-------------------------------------------------------------------------
subroutine read_auxvar_file()
  use variables_module
  implicit none
  integer :: ir,iz,iformat
  double precision :: a,b,c,d,e
  !
  ! Open the file
  !
  open(unit=1,file='auxvar.inp',status='old')
  read(1,*) iformat
  if(iformat.lt.1.or.iformat.gt.1) then
     write(*,*) 'ERROR: iformat of auxvar.inp file not known: ',iformat
     stop
  endif
  read(1,*) ir,iz
  if((ir.ne.grid_nr).or.(iz.ne.grid_nz)) then
     write(*,*) 'ERROR: Dimensions of auxvar.inp file inconsistent: ',ir,iz
     stop
  endif
  !
  ! Allocate the turbalpha array
  !
  allocate(turbalpha(1:grid_nz,1:grid_nr))
  !
  ! Read the stuff
  !
  do ir=1,grid_nr
     do iz=1,grid_nz
        read(1,*) a
        turbalpha(iz,ir) = a
     enddo
  enddo
  close(1)
  !  
end subroutine read_auxvar_file


!-------------------------------------------------------------------------
!                      READ THE dustdens.inp FILE
!-------------------------------------------------------------------------
subroutine read_dustdens_file()
  use variables_module
  implicit none
  integer :: ir,iz,ispec,iformat,is
  double precision, allocatable :: data(:)
  double precision :: a,m
  logical :: fex
  double precision :: pi
  parameter(pi=3.1415926535897932385d0)
  !
  ! Make sure to start with a clean memory
  !
  if(allocated(rhodust)) deallocate(rhodust)
  if(allocated(dust_m)) deallocate(dust_m)
  if(allocated(dust_a)) deallocate(dust_a)
  if(allocated(dust_sigfr)) deallocate(dust_sigfr)
  dust_nrspec = 0
  !
  ! Check if file exists, otherwise no dust
  !
  inquire(file='dustdens.inp',exist=fex)
  if(fex) then 
     !
     ! First get the dust properties
     !
     inquire(file='dustinfo.inp',exist=fex)
     if(.not.fex) then 
        write(*,*) 'ERROR: If you include dust, then also need dustinfo.inp file'
        stop
     endif
     open(unit=1,file='dustinfo.inp',status='old')
     read(1,*) iformat
     if(iformat.lt.1.or.iformat.gt.1) then
        write(*,*) 'ERROR: iformat of dustinfo.inp file not known: ',iformat
        stop
     endif
     read(1,*) dust_nrspec
     if(dust_nrspec.lt.0) stop 7401
     if(dust_nrspec.ge.1) then
        !
        ! Allocate information about the dust masses and sizes
        !
        allocate(dust_m(1:dust_nrspec),dust_a(1:dust_nrspec))
        allocate(dust_sigfr(1:dust_nrspec))
        !
        ! Read this information
        !
        do ispec=1,dust_nrspec
           read(1,*) a,m
           dust_a(ispec) = a
           dust_m(ispec) = m
           dust_sigfr(ispec) = pi * a**2
        enddo
        !
        close(1)
        !
        ! Open the file
        !
        open(unit=1,file='dustdens.inp',status='old')
        read(1,*) iformat
        if(iformat.lt.1.or.iformat.gt.1) then
           write(*,*) 'ERROR: iformat of dustdens.inp file not known: ',iformat
           stop
        endif
        read(1,*) ir,iz
        if((ir.ne.grid_nr).or.(iz.ne.grid_nz)) then
           write(*,*) 'ERROR: Dimensions of dustdens.inp file inconsistent: ',ir,iz
           stop
        endif
        read(1,*) ispec
        if(ispec.ne.dust_nrspec) then
           write(*,*) 'ERROR: The nr of dust species in the file dustdens.inp '
           write(*,*) '       is not the same as those in dustinfo.inp'
           stop
        endif
        !
        ! Allocate the array
        !
        allocate(rhodust(1:grid_nz,1:dust_nrspec,1:grid_nr))
        allocate(data(1:dust_nrspec))
        !
        ! Read the stuff
        !
        do ir=1,grid_nr
           do iz=1,grid_nz
              read(1,*) data
              rhodust(iz,:,ir) = data(:)
              do is=1,dust_nrspec
                 if(data(is).lt.0.d0) then
                    write(*,*) 'ERROR: Negative dust density found in input file...'
                    stop
                 endif
              enddo
           enddo
        enddo
        close(1)
     endif
     !
     ! Free temporary memory
     !
     if(allocated(data)) deallocate(data)
  endif
  !  
end subroutine read_dustdens_file



!-------------------------------------------------------------------------
!                      READ THE chemdens.inp FILE
!-------------------------------------------------------------------------
subroutine read_chemdens_file()
  use variables_module
  implicit none
  integer :: ir,iz,ispec,iformat,is
  double precision, allocatable :: data(:)
  logical :: fex
  !
  ! Make sure to start with a clean memory
  !
  chem_nrspec = 0
  if(allocated(rhochem)) deallocate(rhochem)
  !
  ! Check if file exists, otherwise no gas chemical species
  !
  inquire(file='chemdens.inp',exist=fex)
  if(fex) then 
     !
     ! Open the file
     !
     open(unit=1,file='chemdens.inp',status='old')
     read(1,*) iformat
     if(iformat.lt.1.or.iformat.gt.1) then
        write(*,*) 'ERROR: iformat of chemdens.inp file not known: ',iformat
        stop
     endif
     read(1,*) ir,iz
     if((ir.ne.grid_nr).or.(iz.ne.grid_nz)) then
        write(*,*) 'ERROR: Dimensions of chemdens.inp file inconsistent: ',ir,iz
        stop
     endif
     read(1,*) chem_nrspec
     if(chem_nrspec.lt.0) stop 6650
     if(chem_nrspec.ge.1) then
        !
        ! Allocate the array
        !
        allocate(rhochem(1:grid_nz,1:chem_nrspec,1:grid_nr))
        allocate(data(1:chem_nrspec))
        !
        ! Read the stuff
        !
        do ir=1,grid_nr
           do iz=1,grid_nz
              read(1,*) data
              rhochem(iz,:,ir) = data(:)
              do is=1,chem_nrspec
                 if(data(is).lt.0.d0) then
                    write(*,*) 'ERROR: Negative chem density found in input file...'
                    stop
                 endif
              enddo
           enddo
        enddo
        close(1)
     endif
     !
     ! Free temporary memory
     !
     if(allocated(data)) deallocate(data)
  endif
  !  
end subroutine read_chemdens_file


!-------------------------------------------------------------------------
!                    WRITE THE dustdens_###.out FILE
!-------------------------------------------------------------------------
subroutine write_dustdens_file(itime)
  use variables_module
  implicit none
  integer :: ir,iz,ispec,itime
  double precision, allocatable :: data(:)
  double precision :: a,m
  logical :: fex
  character*80 base,ext,filename
  !
  ! Open the file
  !
  base = 'dustdens_'
  ext  = '.out'
  call make_indexed_filename_filled(base,itime,ext,filename)
  open(unit=1,file=filename)
  write(1,*) 1
  write(1,*) grid_nr,grid_nz
  write(1,*) dust_nrspec
  !
  ! Write the stuff
  !
  allocate(data(1:dust_nrspec))
  do ir=1,grid_nr
     do iz=1,grid_nz
        data(:) = rhodust(iz,:,ir) 
        write(1,*) data
     enddo
  enddo
  close(1)
  !
  ! Free temporary memory
  !
  if(allocated(data)) deallocate(data)
  !  
end subroutine write_dustdens_file


!-------------------------------------------------------------------------
!                   WRITE THE chemdens_###.out FILE
!-------------------------------------------------------------------------
subroutine write_chemdens_file(itime)
  use variables_module
  implicit none
  integer :: ir,iz,ispec,itime
  double precision, allocatable :: data(:)
  double precision :: a,m
  logical :: fex
  character*80 base,ext,filename
  !
  ! Open the file
  !
  base = 'chemdens_'
  ext  = '.out'
  call make_indexed_filename_filled(base,itime,ext,filename)
  open(unit=1,file=filename)
  write(1,*) 1
  write(1,*) grid_nr,grid_nz
  write(1,*) chem_nrspec
  !
  ! Write the stuff
  !
  allocate(data(1:chem_nrspec))
  do ir=1,grid_nr
     do iz=1,grid_nz
        data(:) = rhochem(iz,:,ir) 
        write(1,*) data
     enddo
  enddo
  close(1)
  !
  ! Free temporary memory
  !
  if(allocated(data)) deallocate(data)
  !  
end subroutine write_chemdens_file


!--------------------------------------------------------------------------
!                          DRIFT VELOCITY
!
! This routine computes the drift velocity of a dust particle with respect
! to the gas. As input it needs the external force fext and the
! cross-section to gas drag sigfric.
!
! NOTE: We now use by default the better Epstein drag law, that is
!       consistent with Schaaf (1963) Handbuch der Physik, volume VIII/2, 
!       chapter: Mechanics of Rareﬁed Gases (Springer Verlag, Berlin, 
!       G\"ottingen, Heidelberg), p. 591, and thus with Woitke &
!       Helling (2003) and Tanaka, Himeno & Ida (2005). You can still,
!       for backward compatibility, use the old Dullemond & Dominik (2004)
!       formula, if you set the keyword old.
!--------------------------------------------------------------------------
function vdrift(tgas,rhogas,mugas,fext,sigfric,old)
  implicit none
  doubleprecision :: vdrift,tgas,rhogas,mugas,fext,sigfric
  logical, optional :: old
  !
  doubleprecision :: kk,mp,cc
  parameter(kk  = 1.3807d-16)    ! Bolzmann's constant     [erg/K]
  parameter(mp  = 1.6726d-24)    ! Mass of proton          [g]
  parameter(cc  = 2.9979d10)     ! Light speed             [cm/s]
  doubleprecision :: cs,cst,dum1
  parameter(cst=2.1277d0)        ! 8*sqrt(2)/(3*sqrt(pi))
  logical :: useold
  !
  if(rhogas.lt.1d-60) then
     vdrift = cc                ! A rediculously large velocity
     return
  endif
  useold = .false.
  if(present(old)) then 
     if(old) useold=.true.
  endif
  ! 
  ! Isothermal sound speed
  !
  cs     = sqrt(kk*tgas/(mugas*mp))
  !
  ! Now the actual formula
  !
  if(useold) then
     !
     ! Formula of Dullemond & Dominik 2004b: Eq.(8). These are
     ! not quite accurate. This is an error in that paper.
     !
     vdrift = (3./4.) * fext / (rhogas*sigfric*cs)
  else
     !
     ! Better formula (gives lower drift velocity)
     !
     vdrift = fext / (cst*rhogas*sigfric*cs)
  endif
  !
  ! Limit
  !
  if(vdrift.gt.cc) then
     vdrift=cc
  endif
  return
end function vdrift



!-------------------------------------------------------------------
!                    DO THE TURBULENT DIFFUSION 
!-------------------------------------------------------------------
subroutine do_turb_diff_and_sett(nz,zi,vz,dt,rhogas,turb_nu,f)
  implicit none
  !
  integer nz
  doubleprecision zi(nz+1),f(nz),turb_nu(nz)
  doubleprecision rhogas(nz),vz(nz+1),dt
  doubleprecision fold(nz),dzc(nz),dzi(nz+1)
  doubleprecision dd(nz+1)
  doubleprecision a(nz),b(nz),c(nz)
  doubleprecision rhs(nz),ftold,ftnew,err
  doubleprecision eps,eps1,fact
  integer iz
  doubleprecision ERRTOL
  parameter(ERRTOL=1d-8)
  !
  ! Simple check
  !
  if(vz(1).ne.0.d0) stop 76
  !
  ! Use Crank-Nicholson method, i.e. eps=0.5
  ! (eps=0 --> explicit ; eps=1 --> implicit ; eps=0.5 --> crank-nich)
  !
  eps  = 1.0d0
  eps1 = 1.d0-eps
  !
  !   First make a copy
  !
  do iz=1,nz
     fold(iz) = f(iz)
  enddo
  !
  ! Now make the dz over the cells
  !
  do iz=1,nz
     dzc(iz) = zi(iz+1)-zi(iz)
  enddo
  !
  ! Now make the dz over the interfaces
  !
  do iz=2,nz
     dzi(iz) = 0.5 * ( zi(iz+1)-zi(iz-1) )
  enddo
  dzi(1)    = 0.5 * ( zi(2) - zi(1) )
  dzi(nz+1) = 0.5 * ( zi(nz+1) - zi(nz) )
  !
  ! Make the diffusion coefficient
  !
  do iz=2,nz
     dd(iz) = 0.5*(rhogas(iz)*turb_nu(iz)+            &
                  rhogas(iz-1)*turb_nu(iz-1))
  enddo
  dd(1)    = 0.d0    ! Flux at the inner boundary zero
  dd(nz+1) = 0.d0    ! Flux at the outer boundary zero
  !
  ! Fill the tridiagonal matrix elements a,b,c 
  !
  a(1) = 0.d0
  do iz=2,nz
     a(iz) = dd(iz)/(dzc(iz)*dzi(iz)*rhogas(iz-1))
  enddo
  do iz=1,nz
     b(iz) = -(1.d0/(dzc(iz)*rhogas(iz))) *                &
              ( dd(iz)/dzi(iz) + dd(iz+1)/dzi(iz+1)  )     &
              + vz(iz)/dzc(iz)
  enddo
  c(nz) = 0.d0
  do iz=1,nz-1
     c(iz) = dd(iz+1)/(dzc(iz)*dzi(iz+1)*rhogas(iz+1))     &
             - vz(iz+1)/dzc(iz)
  enddo
  !
  ! Now fill the right hand side of the matrix equation...
  !
  do iz=2,nz-1
     rhs(iz) = eps1*dt*a(iz)*fold(iz-1) +                  &
                    (1.d0+eps1*dt*b(iz))*fold(iz) +        &
                    eps1*dt*c(iz)*fold(iz+1)
  enddo
  rhs(1)  =     (1.d0+eps1*dt*b(1))*fold(1) +              &
                 eps1*dt*c(1)*fold(2)
  rhs(nz) =     eps1*dt*a(nz)*fold(nz-1) +                 &
                 (1.d0+eps1*dt*b(nz))*fold(nz) 
  !
  ! Now compute Mnew = 1-eps*dt*M, which is the matrix we need to
  ! 'invert'.
  !
  do iz=1,nz
     a(iz) = -eps*dt*a(iz)
     b(iz) = 1.d0-eps*dt*b(iz)
     c(iz) = -eps*dt*c(iz)
  enddo
  !
  ! Renormalize all
  !
  do iz=1,nz 
     a(iz)   = a(iz)/b(iz)
     c(iz)   = c(iz)/b(iz)
     rhs(iz) = rhs(iz)/b(iz)
     b(iz)   = 1.d0
  enddo
  !
  ! Now call tridag to solve the matrix equation
  !
  call tridag(a,b,c,rhs,f,nz)
  !
  ! Check for negative densities
  !
  do iz=1,nz
     if(f(iz).lt.0.d0) f(iz)=0.d0
  enddo
  !
  ! Compute error in flux conservation
  !     
  ftold = 0.d0
  ftnew = 0.d0
  do iz=1,nz
     ftold = ftold + fold(iz)*dzc(iz)
     ftnew = ftnew + f(iz)*dzc(iz)
  enddo
  !
  ! Check the error
  !
  err = abs(ftold/ftnew-1.d0)
  if(err.gt.abs(ERRTOL)) then
     write(*,*) 'ERROR IN DIFFUSION ALGORITHM'
     write(*,*) 'Old Ftot = ',ftold
     write(*,*) 'New Ftot = ',ftnew
     write(*,*) 'Error    = ',err
     stop
  endif
  !
  ! If the user wants this (ERRTOL.lt.0) then correct for the error
  !
  if(ERRTOL.lt.0.d0) then
     fact = ftold/ftnew
     do iz=1,nz
        f(iz) = f(iz)*fact
     enddo
  endif
  !
end subroutine do_turb_diff_and_sett


!----------------------------------------------------------------------
!                      MAKE INDEXED FILENAME
!
!     This routine makes filenames of the type dusttemp_004.dat etc.
!----------------------------------------------------------------------
subroutine make_indexed_filename_filled(base,index,ext,filename)
  implicit none
  character*80 base
  integer index
  character*80 ext
  character*80 filename
  character*12 ch
  if((index.lt.0).or.(index.ge.1000)) then
     write(*,*) 'ERROR in make_indexed_filename_filled()'
     stop 729
  endif
  if(index.lt.10) then
     write(ch,11) index
11   format(I1)
     ch='00'//ch(1:1)
  elseif(index.lt.100) then
     write(ch,12) index
12   format(I2)
     ch='0'//ch(1:2)
  elseif(index.lt.1000) then
     write(ch,13) index
13   format(I3)
  endif
  filename = base(1:lnblnk(base))//ch(1:lnblnk(ch))//ext(1:lnblnk(ext))
end subroutine make_indexed_filename_filled




!-------------------------------------------------------------------
!                   FUNCTION: TEST VALIDITY OF NUMBER
!
!     0 = Number is okay
!     1 = Number is INF
!     2 = Number is NAN
!-------------------------------------------------------------------
function number_invalid(a)
  implicit none
  doubleprecision a,b,c
  integer number_invalid
  logical div,sub
  !
  b=a*2.d0
  b=b/2.d0
  c=a-1.d100
  !
  div = (b.eq.a)
  sub = (c.lt.a)
  !
  if(div.and.sub) then
     number_invalid = 0
  elseif(div) then
     number_invalid = 1
  else
     number_invalid = 2
  endif
  !
  return
end function number_invalid


!-------------------------------------------------------------------
!                 NUMERICAL RECIPES ROUTINE TRIDAG()
!-------------------------------------------------------------------
SUBROUTINE tridag(a,b,c,r,u,n)
  INTEGER n,NMAX
  DOUBLEPRECISION a(n),b(n),c(n),r(n),u(n)
  PARAMETER (NMAX=500)
  INTEGER j
  DOUBLEPRECISION bet,gam(NMAX)
  if(n.gt.NMAX) stop 'tridag: too large. Change NMAX.'
  if(b(1).eq.0.) stop 'tridag: rewrite equations'
  bet=b(1)
  u(1)=r(1)/bet
  do j=2,n
     gam(j)=c(j-1)/bet
     bet=b(j)-a(j)*gam(j)
     if(bet.eq.0.) stop 'tridag failed'
     u(j)=(r(j)-a(j)*u(j-1))/bet
  enddo
  do j=n-1,1,-1
     u(j)=u(j)-gam(j+1)*u(j+1)
  enddo
  return
END SUBROUTINE tridag
!  (C) Copr. 1986-92 Numerical Recipes Software =v1.9"217..


